Deadened = LibStub("AceAddon-3.0"):NewAddon("Deadened", "AceHook-3.0", "AceTimer-3.0", "AceConsole-3.0", "AceEvent-3.0", "LibSink-2.0")

local L = LibStub("AceLocale-3.0"):GetLocale("Deadened")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local bit_band = _G.bit.band

local stockAnnounceModes = {None = L["None"], Group = L["Group"], Party = L["Party"], Raid = L["Raid"], Say = L["Say"], Yell = L["Yell"], RaidWarn = L["Raid Warning"], Battleground = L["Battleground"]}
local announceModes = {}
local db

local options = {
	type = "group",
	args = {
		spells = {
			type = "group",
			name = L["Spells to suppress"],
			desc = L["Select the casts you want to suppress"],
			args = {
				add = {
					type = "input",
					name = L["Add a spell ID"],
					desc = L["Add the spell ID of a spell to suppress (must be numeric)"],
					get = function() return "" end,
					set = function(info, v)
						db.spells[v] = true
						Deadened:AddSpell(v, true)
						Deadened:RebuildSpellTable()
					end,
					pattern = "%d+"
				}
			}
		},
		announceSpells = {
			type = "group",
			name = L["Spells to announce"],
			desc = L["Spells to announce hits/misses for"],
			args = {
				add = {
					type = "input",
					name = L["Add a spell ID/name"],
					desc = L["Add the spell ID or spell name of a spell to announce"],
					get = function() return "" end,
					set = function(info, v)
						local link = GetSpellLink(v)
						if link then
							local id = link:match("Hspell:(%d+)")
							if id then
								db.announceSpells[id] = true
								Deadened:AddAnnounceSpell(id, true)
								Deadened:RebuildSpellTable()
							end							
						end
					end
				}
			}
		},
		announceRaid = {
			type = "select",
			name = L["Announce (Raid-25)"],
			desc = L["Announce while in a 25-man raid to..."],
			get = function()
				return db.announceRaid or db.announce
			end,
			set = function(info, v)
				db.announceRaid = v
				Deadened:Print(L["Now announcing 25-man raid interrupts to"], v)
			end,
			values = announceModes
		},
		announceRaid10 = {
			type = "select",
			name = L["Announce (Raid-10)"],
			desc = L["Announce while in a 10-man raid to..."],
			get = function()
				return db.announceRaid10 or db.announceRaid or db.announce
			end,
			set = function(info, v)
				db.announceRaid10 = v
				Deadened:Print(L["Now announcing 10-man raid interrupts to"], v)
			end,
			values = announceModes
		},		
		announceArena = {
			type = "select",
			name = L["Announce (Arena)"],
			desc = L["Announce while in an arena to..."],
			get = function()
				return db.announceArena or db.announce
			end,
			set = function(info, v)
				db.announceArena = v
				Deadened:Print(L["Now announcing arena interrupts to"], v)
			end,
			values = announceModes
		},	
		announceParty = {
			type = "select",
			name = L["Announce (Party)"],
			desc = L["Announce while in a party to..."],
			get = function()
				return db.announceParty or db.announce
			end,
			set = function(info, v)
				db.announceParty = v
				Deadened:Print(L["Now announcing party interrupts to"], v)
			end,
			values = announceModes
		},
		announceBattleground = {
			type = "select",
			name = L["Announce (Battleground)"],
			desc = L["Announce while in a battleground to..."],
			get = function()
				return db.announceBattleground or db.announce
			end,
			set = function(info, v)
				db.announceBattleground = v
				Deadened:Print(L["Now announcing battleground interrupts to"], v)
			end,
			values = announceModes
		},
		announceOther = {
			type = "select",
			name = L["Announce (Other)"],
			desc = L["Announce to..."],
			get = function()
				return db.announceOther or db.announce
			end,
			set = function(info, v)
				db.announceOther = v
				Deadened:Print(L["Now announcing other interrupts to"], v)
			end,
			values = announceModes
		},
		announceMisses = {
			type = "toggle",
			name = L["Announce misses"],
			get = function()
				return db.announceMisses
			end,
			set = function(info, v)
				db.announceMisses = v
			end
		},
		modules = {
			type = "group",
			name = L["Modules"],
			desc = L["Modules"],
			args = {}
		}
	}
}
	
AceConfig:RegisterOptionsTable("Deadened", options, {"deaden", "deadened"})
Deadened:SetDefaultModuleState(false)

local interrupts = {
	[1766] = "Kick",
	[47528] = "Mind Freeze",
	[6552] = "Pummel",
	[72] = "Shield Bash",
	[2139] = "Counterspell"
}

function Deadened:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("DeadenedDB")
	self.db:RegisterDefaults({
		profile = {
			modules = {},
			tankmode = false,
			sink20OutputSink = "None",
			announce = "None",
			announceMisses = true,
			spells = {
				[41426] = false,
				[41410] = true,
				[40832] = false,
				[40904] = false,
				-- [13878] = true 		-- Scorch, testing
			},
			announceSpells = {
				[2094] = false
			}
		}
	})
	db = self.db.profile
	optFrame = AceConfigDialog:AddToBlizOptions("Deadened", "Deadened")
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	
	for k, v in self:IterateModules() do
		options.args.modules.args[k] = {
			type = "toggle", 
			name = k, 
			desc = L["Toggle "] .. k, 
			get = function()
				return db.modules[k] ~= false or false
			end,
			set = function(info, v)
				db.modules[k] = v
				if v then
					Deadened:EnableModule(k)
					Deadened:Print(L["Enabled"], k, L["module"])
				else
					Deadened:DisableModule(k)
					Deadened:Print(L["Disabled"], k, L["module"])
				end
			end,
			hidden = function() return v.MenuHide and v:MenuHide() end
		}
	end		
	
end

local serverChannels = {}
local function excludeChannels(...)
	for i = 1, select("#", ...) do
		local name = select(i, ...)
		serverChannels[name] = true
	end
end

function Deadened:AddCustomChannels(...)
	for k, v in pairs(announceModes) do
		announceModes[k] = nil
	end
	for k, v in pairs(stockAnnounceModes) do
		announceModes[k] = v
	end
	excludeChannels(EnumerateServerChannels())
	for i = 1, select("#", ...), 2 do
		local id, name = select(i, ...)
		if not serverChannels[name] then
			announceModes[name] = name
		end
	end
end

local spellNames, announceSpellNames = {}, {}

function Deadened:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE")
	for k, v in pairs(db.spells) do
		self:AddSpell(k, v)
	end
	for k, v in pairs(db.announceSpells) do
		self:AddAnnounceSpell(k, v)
	end
	
	for k, v in self:IterateModules() do
		if db.modules[k] ~= false then
			v:Enable()
		end
	end
	self:AddCustomChannels(GetChannelList())
end

function Deadened:CHAT_MSG_CHANNEL_NOTICE()
	self:AddCustomChannels(GetChannelList())
end

function Deadened:AddSpell(k, v)
	k = tonumber(k)
	local spellName = GetSpellInfo(k)
	spellNames[spellName] = v
	options.args.spells.args[spellName:gsub(" ", "")] = {
		type = "toggle",
		name = spellName,
		desc = spellName,
		get = function(info) return db.spells[k] end,
		set = function(info, val)
			db.spells[k] = val
			Deadened:RebuildSpellTable()
		end,
	}
end

function Deadened:AddAnnounceSpell(k, v)
	k = tonumber(k)
	local spellName = GetSpellInfo(k)
	self:Print("Adding", spellName, "for announce")
	announceSpellNames[spellName] = v
	options.args.announceSpells.args[spellName:gsub(" ", "")] = {
		type = "toggle",
		name = spellName,
		desc = spellName,
		get = function(info) return db.announceSpells[k] end,
		set = function(info, val)
			db.announceSpells[k] = val
			Deadened:RebuildSpellTable()
		end,
	}
end

function Deadened:RebuildSpellTable()
	for k, v in pairs(db.spells) do
		local spellName = GetSpellInfo(k)
		spellNames[spellName] = v
	end

	for k, v in pairs(db.announceSpells) do
		local spellName = GetSpellInfo(k)
		announceSpellNames[spellName] = v
	end
end

function Deadened:InitModule(k)
	if db.modules[k] ~= false then
		self:DisableModule(k)
		self:EnableModule(k)
	end
end

function Deadened:Dump()
	for k, v in pairs(spellNames) do
		self:Print(k, v)
	end
end

local function getDistribution(p)
	if p:lower() == "party" then
		if GetNumPartyMembers() > 0 then return "PARTY" end
	elseif p:lower() == "raid" then
		if GetNumRaidMembers() > 0 then return "RAID" end
	elseif p:lower() == "raidwarn" then
		if GetNumRaidMembers() > 0 then return "RAID_WARNING" end
	elseif p:lower() == "group" then
		if GetNumRaidMembers() > 0 then return "RAID" end
		if GetNumPartyMembers() > 0 then return "PARTY" end
	elseif p:lower() == "say" then
		return "SAY"
	elseif p:lower() == "yell" then
		return "YELL"
	elseif p then
		return "CHANNEL"
	end
	return nil
end

local function getDistributionSetting()
	local inInstance, instanceType = IsInInstance()
	if instanceType == "arena" then
		return db.announceArena
	elseif instanceType == "pvp" then	
		return db.announceBattleground
	elseif GetNumRaidMembers() > 10 then
		return db.announceRaid
	elseif GetNumRaidMembers() > 0 then
		return db.announceRaid10
	elseif GetNumPartyMembers() > 0 then
		return db.announceParty
	else
		return db.announceOther
	end
end

local COMBATLOG_FILTER_ME = _G.COMBATLOG_FILTER_ME
local immunityIDs = {
	642, 498, 1022, 5599, 10278, 45438
}
local immunities = {}
for _, id in ipairs(immunityIDs) do
	local name = GetSpellInfo(id)
	immunities[name] = true
end
do
	local function getMissReason(unit)
		for k, v in pairs(immunities) do
			local name, _, _, _, _, _, duration = UnitBuff(unit, k)
			duration = duration or "(unknown)"
			if name then
				return name, duration
			end
		end
		return nil, nil
	end
	
	local function sendMessage(msg)
		local setting = getDistributionSetting() or db.announce
		if setting == "None" then return end
		local d = getDistribution(setting)
		if d == "CHANNEL" then
			SendChatMessage(msg, d, nil, GetChannelName(setting))
		elseif d then
			SendChatMessage(msg, d)
		end	
	end
	
	local symbols = {"{star}", "{circle}", "{diamond}", "{triangle}", "{moon}", "{square}", "{cross}", "{skull}"}
	local lastGUID, lastTime
	local GetTime = _G.GetTime
	local UnitGUID = _G.UnitGUID
	local UnitExists = _G.UnitExists

	local function getUnit(guid)
		return (UnitGUID("target") and "target") or (guid == UnitGUID("focus") and "focus") or (guid == UnitGUID("mouseover") and "mouseover")
	end

	function Deadened:COMBAT_LOG_EVENT_UNFILTERED(event, ...)
		local timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellID, spellName, _, missType = ...
		if eventtype == "SPELL_INTERRUPT" and srcGUID == UnitGUID("player") then
			local srcSpellId, srcSpellName, srcSpellSchool, dstSpellId, dstSpellName, dstSpellSchool = select(9, ...);

			-- This prevents problems with Throwing Weapon Spec reporting 2/4 interrupts
			local skip = dstGUID == lastGUID and GetTime() - lastTime < 0.5
			lastGUID = dstGUID
			lastTime = GetTime()
			
			local symbol
			local unit = getUnit(dstGUID)
			if unit then
				symbol = symbols[GetRaidTargetIndex(unit) or 0]
			end
			if not skip then
				local msg = ("%sed %s%s's %s!"):format(srcSpellName, symbol or "", dstName, dstSpellName)
				sendMessage(msg)
			end	
		elseif eventtype == "SPELL_CAST_SUCCESS" and srcGUID == UnitGUID("player") and interrupts[spellID] then
			local msg = ("%s %s %s %s!"):format(L["Used"], spellName, L["on"], dstName)
			local unit = getUnit(dstGUID)
			if unit then
				local spell, rank, displayName, icon, startTime, endTime = UnitCastingInfo(unit)
				if not spell then
					sendMessage(msg)
				end
			end
		elseif eventtype == "SPELL_AURA_APPLIED" and srcGUID == UnitGUID("player") and announceSpellNames[spellName] then
			msg = ("%s %s %s!"):format(spellName, L["is up on"], dstName)
			sendMessage(msg)
		elseif eventtype == "SPELL_MISSED" and srcGUID == UnitGUID("player") and (interrupts[spellID] or announceSpellNames[spellName]) then
			local unit = getUnit(dstGUID)
			local reason, timeleft, msg
			if missType == "IMMUNE" then
				reason, timeleft = getMissReason(unit)
			end
			if reason then
				self:Print(reason, timeleft)
				msg = (L["Missed %s on %s (%s - %s, %s seconds remaining)"]):format(spellName, dstName, missType, reason, timeleft or -1)
			else
				msg = (L["Missed %s on %s (%s)"]):format(spellName, dstName, missType)
			end
			sendMessage(msg)
		end
	end
end

function Deadened:OnDisable()
end

function Deadened:IsDeaden(v)
	if type(v) == "number" then
		return db.spells[v]
	end
	return spellNames[v]
end

Deadened.Suppress = Deadened.IsDeaden

local timer = nil
function Deadened:Alert(s)
	if timer then
		self:CancelTimer(timer, true)
		timer = nil
	end
	timer = self:ScheduleTimer(self.Pour, 0.25, self, s)
end

function Deadened:UNIT_SPELLCAST_START(orig_object, unit, ...)
	if unit == "target" then
		local spell, rank, displayName, icon, startTime, endTime = UnitCastingInfo(unit)
		if not Deadened:IsDeaden(displayName) then
			for k,v in pairs(self.hooks[orig_object]) do
				v(orig_object, unit, ...)
			end
		else
			self:Alert((L["%s casting %s (%s) sec"]):format(UnitName(unit), displayName, ("%2.1f"):format((endTime-startTime) / 1000 )))
		end
	end	
end
